/*
 * File         : mpi_Request.c
 * Headerfile   : mpi_Request.h 
 * Author       : Sung-Hoon Ko, Xinying Li, Bryan Carpenter
 * Created      : Thu Apr  9 12:22:15 1998
 * Revision     : $Revision: 1.3 $
 * Updated      : $Date: 1999/09/02 17:31:45 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include <mpi.h>
#include "mpi_Request.h"

jfieldID reqhandleID;
extern jfieldID stathandleID;
extern jfieldID sourceID,tagID,countID, indexID;

/*
 * Class:     mpi_Status
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Request_init 
  (JNIEnv *env, jclass thisClass)
{
  reqhandleID = (*env)->GetFieldID(env,thisClass,"handle","J");
}                      

/*
 * Class:     mpi_Request
 * Method:    GetReq
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_mpi_Request_GetReq
  (JNIEnv *env, jobject jthis, jint type)
{
  switch (type) {
    case 0:
      (*env)->SetLongField(env,jthis,reqhandleID,(jlong)MPI_REQUEST_NULL);
      break;
	default:
	  break;
  }
}

/*
 * Class:     mpi_Request
 * Method:    Cancel
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Request_Cancel
  (JNIEnv *env, jobject jthis)
{
   MPI_Request req;
   req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
   MPI_Cancel(&req);
}

/*
 * Class:     mpi_Request
 * Method:    Free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_mpi_Request_Free
  (JNIEnv *env, jobject jthis)
{
   MPI_Request req;
   req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
   MPI_Request_free(&req);
}

/*
 * Class:     mpi_Request
 * Method:    Is_null
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_mpi_Request_Is_1null
  (JNIEnv *env, jobject jthis)
{
  MPI_Request req;
  req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
  if(req==MPI_REQUEST_NULL)
    return JNI_TRUE;
  else 
    return JNI_FALSE;
}

/*
 * Class:     mpi_Request
 * Method:    Wait
 * Signature: (Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Wait
  (JNIEnv *env, jobject jthis, jobject stat)
{
  MPI_Request req;
  MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
  jclass status_class = (*env)->GetObjectClass(env, stat);
  sourceID=(*env)->GetFieldID(env,status_class,"source","I");
  tagID=(*env)->GetFieldID(env,status_class,"tag","I");
*/

  req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));

  MPI_Wait(&req, status);

  (*env)->SetLongField(env,jthis,reqhandleID,(jlong)req);

  (*env)->SetLongField(env,stat,stathandleID,(jlong)status);
  (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
  (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);

  return stat;
}

/*
 * Class:     mpi_Request
 * Method:    Test
 * Signature: (Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Test
  (JNIEnv *env, jobject jthis, jobject stat)
{
  int flag;
  MPI_Request req;
  MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
  jclass status_class = (*env)->GetObjectClass(env, stat);
  sourceID=(*env)->GetFieldID(env,status_class,"source","I");
  tagID=(*env)->GetFieldID(env,status_class,"tag","I");
*/

  req=(MPI_Request)((*env)->GetLongField(env,jthis,reqhandleID));
  
  MPI_Test(&req, &flag, status);

  (*env)->SetLongField(env,jthis,reqhandleID,(jlong)req);

  (*env)->SetLongField(env,stat, stathandleID,(jlong) status);
  (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
  (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);

  if(flag!=0)
    return stat;
  else {
    free(status) ;
    return NULL;
  }
}

/*
 * Class:     mpi_Request
 * Method:    Waitany
 * Signature: ([Lmpi/Request;Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Waitany
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request, jobject stat)
{
  int i,index;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(count, sizeof(MPI_Request));
  MPI_Status *status = (MPI_Status*)malloc(sizeof(MPI_Status));

/*
  jclass status_class = (*env)->GetObjectClass(env, stat);
  sourceID=(*env)->GetFieldID(env,status_class,"source","I");
  tagID=(*env)->GetFieldID(env,status_class,"tag","I");
  indexID=(*env)->GetFieldID(env,status_class,"index","I");
*/

  for(i=0; i<count; i++) {
    reqs[i]=(MPI_Request)( (*env)->GetLongField(env,
	  (*env)->GetObjectArrayElement(env,array_of_request,i),
          reqhandleID) );
   }

  MPI_Waitany(count, reqs, &index, status);

  for(i=0; i<count; i++) {
    (*env)->SetLongField( env,
              (*env)->GetObjectArrayElement(env,array_of_request,i),
              reqhandleID, (jlong) reqs[i] );
	}

  (*env)->SetLongField(env,stat, stathandleID,(jlong) status); /* shko */

  (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
  (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);
  (*env)->SetIntField(env,stat, indexID,index);

  free(reqs);

  return stat;
}

/*
 * Class:     mpi_Request
 * Method:    Testany
 * Signature: ([Lmpi/Request;Lmpi/Status;)Lmpi/Status;
 */
JNIEXPORT jobject JNICALL Java_mpi_Request_Testany
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request, jobject stat)
{
  int i,flag,index;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(count, sizeof(MPI_Request));
  MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
  jclass status_class = (*env)->GetObjectClass(env, stat);
  sourceID=(*env)->GetFieldID(env,status_class,"source","I");
  tagID=(*env)->GetFieldID(env,status_class,"tag","I");
  indexID=(*env)->GetFieldID(env,status_class,"index","I");
*/

  for(i=0; i<count; i++) {
    reqs[i]=(MPI_Request)( (*env)->GetLongField(env,
        (*env)->GetObjectArrayElement(env,array_of_request,i),
        reqhandleID) );
  }

  MPI_Testany(count, reqs, &index,&flag, status);
  
  for(i=0; i<count; i++) {
    (*env)->SetLongField( env,
                (*env)->GetObjectArrayElement(env,array_of_request,i),
                reqhandleID, (jlong) reqs[i] );
  }

  (*env)->SetLongField(env,stat, stathandleID,(jlong) status);
  (*env)->SetIntField(env,stat, sourceID,status->MPI_SOURCE);
  (*env)->SetIntField(env,stat, tagID,status->MPI_TAG);
  (*env)->SetIntField(env,stat, indexID,index);

  free(reqs);
  if(flag!=0)
    return stat;
  else {
    free(status) ; 
    return NULL;
  }
}

/*
 * Class:     mpi_Request
 * Method:    waitall
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_waitall
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(count, sizeof(MPI_Request));
  MPI_Status *stas=(MPI_Status*)calloc(count, sizeof(MPI_Status));

  jclass status_class=(*env)->FindClass(env,"mpi/Status");

  jmethodID handleConstructorID = 
    (*env)->GetMethodID(env, status_class, "<init>", "()V");

  jobjectArray array_of_status = 
    (*env)->NewObjectArray(env,count,status_class,NULL);

  for(i=0; i<count; i++) {
    reqs[i]=(MPI_Request)((*env)->GetLongField( env,
        (*env)->GetObjectArrayElement(env,array_of_request,i),
        reqhandleID) );
  }

  MPI_Waitall(count, reqs, stas);

  for(i=0; i<count; i++) {
    (*env)->SetLongField( env,
             (*env)->GetObjectArrayElement(env,array_of_request,i),
             reqhandleID, (jlong) reqs[i] );
  }
    
  for(i=0; i<count; i++) { 
    jobject jstas=(*env)->NewObject(env,status_class,handleConstructorID);
    MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
    sourceID=(*env)->GetFieldID(env,status_class,"source","I");
    tagID=(*env)->GetFieldID(env,status_class,"tag","I");
*/

    *status = stas [i] ;
    (*env)->SetLongField(env,jstas,stathandleID,(jlong) status);
    (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
    (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   

    (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);
  }

  free(reqs);
  free(stas);

  return array_of_status;
}


/*
 * Class:     mpi_Request
 * Method:    testall
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_testall
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i,flag;
  int count=(*env)->GetArrayLength(env,array_of_request);
  MPI_Status *stas=(MPI_Status*)calloc(count, sizeof(MPI_Status));
  MPI_Request *reqs=(MPI_Request*)calloc(count, sizeof(MPI_Request));

  jclass status_class=(*env)->FindClass(env,"mpi/Status");

  jmethodID handleConstructorID = 
    (*env)->GetMethodID(env, status_class, "<init>", "()V");

  jobjectArray array_of_status = 
    (*env)->NewObjectArray(env,count,status_class,NULL);

  for(i=0; i<count; i++) {
    reqs[i]=( MPI_Request)((*env)->GetLongField(env,
        (*env)->GetObjectArrayElement(env,array_of_request,i),
        reqhandleID) );
  }

  MPI_Testall(count, reqs, &flag, stas);

  free(reqs);

  for(i=0; i<count; i++)
    (*env)->SetLongField( env,
		(*env)->GetObjectArrayElement(env,array_of_request,i),
                reqhandleID, (jlong) reqs[i] );

  if(flag!=0) {
    for(i=0; i<count; i++) {
      jobject jstas=(*env)->NewObject(env,status_class,handleConstructorID);
      MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
      sourceID=(*env)->GetFieldID(env,status_class,"source","I");
      tagID=(*env)->GetFieldID(env,status_class,"tag","I");
*/

      *status = stas [i] ;
      (*env)->SetLongField(env,jstas,stathandleID,(jlong) status);
      (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
      (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);  

      (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);
    }
  }

  free(stas);

  if(flag!=0)
    return array_of_status;
  else
    return NULL;
}

/*
 * Class:     mpi_Request
 * Method:    waitsome
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_waitsome
  (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i;
  int incount=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(incount, sizeof(MPI_Request));
  MPI_Status *stas=(MPI_Status*)calloc(incount, sizeof(MPI_Status));
  int *array_of_indices=(int*)calloc(incount,sizeof(int));
  int outcount;

  jclass status_class=(*env)->FindClass(env,"mpi/Status");
  jmethodID handleConstructorID = 
    (*env)->GetMethodID(env, status_class, "<init>", "()V");

  jobjectArray array_of_status;

  for(i=0; i<incount; i++) {
    reqs[i] = 
      (MPI_Request)( (*env)->GetLongField(env,
	   (*env)->GetObjectArrayElement(env,array_of_request,i),
           reqhandleID) );
  }

  MPI_Waitsome(incount, reqs, &outcount, array_of_indices, stas); 

  if(outcount!=MPI_UNDEFINED) {
    for(i=0; i<outcount; i++) {
      (*env)->SetLongField( env,
	          (*env)->GetObjectArrayElement( env,
                  array_of_request,array_of_indices[i] ),
                  reqhandleID,(jlong) reqs[array_of_indices[i]] );
    }

    array_of_status=(*env)->NewObjectArray(env,outcount,status_class,NULL);

    for(i=0; i<outcount; i++) {
      jobject jstas=(*env)->NewObject(env,status_class,handleConstructorID);
      MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
	sourceID=(*env)->GetFieldID(env,status_class,"source","I");
	tagID=(*env)->GetFieldID(env,status_class,"tag","I");
	indexID=(*env)->GetFieldID(env,status_class,"index","I");
*/

      *status = stas [i] ;
      (*env)->SetLongField(env,jstas,stathandleID,(jlong) status);
      (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
      (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   
      (*env)->SetIntField(env,jstas,indexID,array_of_indices[i]);

      (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);
    }
  }

  free(reqs);
  free(stas);
  free(array_of_indices);

  if(outcount==MPI_UNDEFINED)
    return NULL;
  else
    return array_of_status;
}

/*
 * Class:     mpi_Request
 * Method:    testsome
 * Signature: ([Lmpi/Request;)[Lmpi/Status;
 */
JNIEXPORT jobjectArray JNICALL Java_mpi_Request_testsome
 (JNIEnv *env, jclass jthis, jobjectArray array_of_request)
{
  int i;
  int incount=(*env)->GetArrayLength(env,array_of_request);
  MPI_Request *reqs=(MPI_Request*)calloc(incount, sizeof(MPI_Request));
  MPI_Status *stas=(MPI_Status*)calloc(incount, sizeof(MPI_Status));
  int *array_of_indices=(int*)calloc(incount,sizeof(int));
  int outcount;

  jclass status_class=(*env)->FindClass(env,"mpi/Status");

  jmethodID handleConstructorID = 
    (*env)->GetMethodID(env, status_class, "<init>", "()V");

  jobjectArray array_of_status;

  for(i=0; i<incount; i++) {
    reqs[i]=(MPI_Request)( (*env)->GetLongField(env,
        (*env)->GetObjectArrayElement(env,array_of_request,i),
        reqhandleID) );
  }

  MPI_Testsome(incount,reqs,&outcount,array_of_indices, stas);

  if(outcount!=MPI_UNDEFINED) {  
    for(i=0; i<outcount; i++) {
      (*env)->SetLongField( env,
		   (*env)->GetObjectArrayElement(env,
                   array_of_request,array_of_indices[i]),
		   reqhandleID,
                   (jlong) reqs[array_of_indices[i]] );
    }

    array_of_status = (*env)->NewObjectArray(env,outcount,status_class,NULL);
    for(i=0; i<outcount; i++) {
      jobject jstas=(*env)->NewObject(env,status_class,handleConstructorID);
      MPI_Status *status = (MPI_Status*) malloc(sizeof(MPI_Status));

/*
      sourceID=(*env)->GetFieldID(env,status_class,"source","I");
      tagID=(*env)->GetFieldID(env,status_class,"tag","I");
      indexID=(*env)->GetFieldID(env,status_class,"index","I");
*/

      *status = stas [i] ;
      (*env)->SetLongField(env,jstas,stathandleID,(jlong) status);
      (*env)->SetIntField(env,jstas,sourceID,status->MPI_SOURCE);
      (*env)->SetIntField(env,jstas,tagID,status->MPI_TAG);   
      (*env)->SetIntField(env,jstas,indexID,array_of_indices[i]);

      (*env)->SetObjectArrayElement(env,array_of_status,i,jstas);
    }
  }

  free(reqs);
  free(stas);
  free(array_of_indices);
  
  if(outcount==MPI_UNDEFINED)
    return NULL;
  else
    return array_of_status;
}

/*
 * Things to do:
 *
 *    Initialize `handleConstructorID' in consistent way with `sourceId', etc.
 *
 */

